@@ -1,5 +1,12 @@
-#ifndef NGX_HTTP_LUA_OUTPUT_H
-#define NGX_HTTP_LUA_OUTPUT_H
+
+/*
+ * Copyright (C) Xiaozhe Wang (chaoslawful)
+ * Copyright (C) Yichun Zhang (agentzh)
+ */
+
+
+#ifndef _NGX_HTTP_LUA_OUTPUT_H_INCLUDED_
+#define _NGX_HTTP_LUA_OUTPUT_H_INCLUDED_
 
 
 #include "ngx_http_lua_common.h"
@@ -16,5 +23,6 @@ ngx_int_t ngx_http_lua_flush_resume_helper(ngx_http_request_t *r,
     ngx_http_lua_ctx_t *ctx);
 
 
-#endif /* NGX_HTTP_LUA_OUTPUT_H */
+#endif /* _NGX_HTTP_LUA_OUTPUT_H_INCLUDED_ */
 
+/* vi:set ft=c ts=4 sw=4 et fdm=marker: */
