@@ -1,3 +1,9 @@
+
+/*
+ * Copyright (C) Yichun Zhang (agentzh)
+ */
+
+
 #ifndef DDEBUG
 #define DDEBUG 0
 #endif
@@ -10,7 +16,10 @@
 
 
 static int ngx_http_lua_shdict_set(lua_State *L);
+static int ngx_http_lua_shdict_safe_set(lua_State *L);
 static int ngx_http_lua_shdict_get(lua_State *L);
+static int ngx_http_lua_shdict_get_stale(lua_State *L);
+static int ngx_http_lua_shdict_get_helper(lua_State *L, int get_stale);
 static int ngx_http_lua_shdict_expire(ngx_http_lua_shdict_ctx_t *ctx,
     ngx_uint_t n);
 static ngx_int_t ngx_http_lua_shdict_lookup(ngx_shm_zone_t *shm_zone,
@@ -18,6 +27,7 @@ static ngx_int_t ngx_http_lua_shdict_lookup(ngx_shm_zone_t *shm_zone,
     ngx_http_lua_shdict_node_t **sdp);
 static int ngx_http_lua_shdict_set_helper(lua_State *L, int flags);
 static int ngx_http_lua_shdict_add(lua_State *L);
+static int ngx_http_lua_shdict_safe_add(lua_State *L);
 static int ngx_http_lua_shdict_replace(lua_State *L);
 static int ngx_http_lua_shdict_incr(lua_State *L);
 static int ngx_http_lua_shdict_delete(lua_State *L);
@@ -28,6 +38,7 @@ static int ngx_http_lua_shdict_get_keys(lua_State *L);
 
 #define NGX_HTTP_LUA_SHDICT_ADD         0x0001
 #define NGX_HTTP_LUA_SHDICT_REPLACE     0x0002
+#define NGX_HTTP_LUA_SHDICT_SAFE_STORE  0x0004
 
 
 ngx_int_t
@@ -124,8 +135,8 @@ ngx_http_lua_shdict_rbtree_insert_value(ngx_rbtree_node_t *temp,
             sdn = (ngx_http_lua_shdict_node_t *) &node->color;
             sdnt = (ngx_http_lua_shdict_node_t *) &temp->color;
 
-            p = (ngx_memn2cmp(sdn->data, sdnt->data, sdn->key_len,
-                sdnt->key_len) < 0) ? &temp->left : &temp->right;
+            p = ngx_memn2cmp(sdn->data, sdnt->data, sdn->key_len,
+                             sdnt->key_len) < 0 ? &temp->left : &temp->right;
         }
 
         if (*p == sentinel) {
@@ -282,17 +293,26 @@ ngx_http_lua_inject_shdict_api(ngx_http_lua_main_conf_t *lmcf, lua_State *L)
         lua_createtable(L, 0, lmcf->shm_zones->nelts /* nrec */);
                 /* ngx.shared */
 
-        lua_createtable(L, 0 /* narr */, 10 /* nrec */); /* shared mt */
+        lua_createtable(L, 0 /* narr */, 13 /* nrec */); /* shared mt */
 
         lua_pushcfunction(L, ngx_http_lua_shdict_get);
         lua_setfield(L, -2, "get");
 
+        lua_pushcfunction(L, ngx_http_lua_shdict_get_stale);
+        lua_setfield(L, -2, "get_stale");
+
         lua_pushcfunction(L, ngx_http_lua_shdict_set);
         lua_setfield(L, -2, "set");
 
+        lua_pushcfunction(L, ngx_http_lua_shdict_safe_set);
+        lua_setfield(L, -2, "safe_set");
+
         lua_pushcfunction(L, ngx_http_lua_shdict_add);
         lua_setfield(L, -2, "add");
 
+        lua_pushcfunction(L, ngx_http_lua_shdict_safe_add);
+        lua_setfield(L, -2, "safe_add");
+
         lua_pushcfunction(L, ngx_http_lua_shdict_replace);
         lua_setfield(L, -2, "replace");
 
@@ -340,6 +360,20 @@ ngx_http_lua_inject_shdict_api(ngx_http_lua_main_conf_t *lmcf, lua_State *L)
 
 static int
 ngx_http_lua_shdict_get(lua_State *L)
+{
+  return ngx_http_lua_shdict_get_helper(L, 0 /* stale */);
+}
+
+
+static int
+ngx_http_lua_shdict_get_stale(lua_State *L)
+{
+  return ngx_http_lua_shdict_get_helper(L, 1 /* stale */);
+}
+
+
+static int
+ngx_http_lua_shdict_get_helper(lua_State *L, int get_stale)
 {
     int                          n;
     ngx_str_t                    name;
@@ -350,7 +384,7 @@ ngx_http_lua_shdict_get(lua_State *L)
     ngx_http_lua_shdict_node_t  *sd;
     ngx_str_t                    value;
     int                          value_type;
-    lua_Number                   num;
+    double                       num;
     u_char                       c;
     ngx_shm_zone_t              *zone;
     uint32_t                     user_flags = 0;
@@ -359,31 +393,35 @@ ngx_http_lua_shdict_get(lua_State *L)
 
     if (n != 2) {
         return luaL_error(L, "expecting exactly two arguments, "
-                "but only seen %d", n);
+                          "but only seen %d", n);
     }
 
-    luaL_checktype(L, 1, LUA_TLIGHTUSERDATA);
-
     zone = lua_touserdata(L, 1);
     if (zone == NULL) {
-        return luaL_error(L, "bad user data for the ngx_shm_zone_t pointer");
+        return luaL_error(L, "bad \"zone\" argument");
     }
 
     ctx = zone->data;
-
     name = ctx->name;
 
+    if (lua_isnil(L, 2)) {
+        lua_pushnil(L);
+        lua_pushliteral(L, "nil key");
+        return 2;
+    }
+
     key.data = (u_char *) luaL_checklstring(L, 2, &key.len);
 
     if (key.len == 0) {
         lua_pushnil(L);
-        return 1;
+        lua_pushliteral(L, "empty key");
+        return 2;
     }
 
     if (key.len > 65535) {
-        return luaL_error(L,
-                          "the key argument is more than 65535 bytes: \"%s\"",
-                          key.data);
+        lua_pushnil(L);
+        lua_pushliteral(L, "key too long");
+        return 2;
     }
 
     hash = ngx_crc32_short(key.data, key.len);
@@ -396,20 +434,22 @@ ngx_http_lua_shdict_get(lua_State *L)
     ngx_shmtx_lock(&ctx->shpool->mutex);
 
 #if 1
-    ngx_http_lua_shdict_expire(ctx, 1);
+    if (!get_stale) {
+        ngx_http_lua_shdict_expire(ctx, 1);
+    }
 #endif
 
     rc = ngx_http_lua_shdict_lookup(zone, hash, key.data, key.len, &sd);
 
     dd("shdict lookup returns %d", (int) rc);
 
-    if (rc == NGX_DECLINED || rc == NGX_DONE) {
+    if (rc == NGX_DECLINED || (rc == NGX_DONE && !get_stale)) {
         ngx_shmtx_unlock(&ctx->shpool->mutex);
         lua_pushnil(L);
         return 1;
     }
 
-    /* rc == NGX_OK */
+    /* rc == NGX_OK || (rc == NGX_DONE && get_stale) */
 
     value_type = sd->value_type;
 
@@ -427,16 +467,16 @@ ngx_http_lua_shdict_get(lua_State *L)
 
     case LUA_TNUMBER:
 
-        if (value.len != sizeof(lua_Number)) {
+        if (value.len != sizeof(double)) {
 
             ngx_shmtx_unlock(&ctx->shpool->mutex);
 
             return luaL_error(L, "bad lua number value size found for key %s "
-                    "in shared_dict %s: %lu", key.data, name.data,
-                    (unsigned long) value.len);
+                              "in shared_dict %s: %lu", key.data, name.data,
+                              (unsigned long) value.len);
         }
 
-        num = *(lua_Number *) value.data;
+        num = *(double *) value.data;
 
         lua_pushnumber(L, num);
         break;
@@ -448,8 +488,8 @@ ngx_http_lua_shdict_get(lua_State *L)
             ngx_shmtx_unlock(&ctx->shpool->mutex);
 
             return luaL_error(L, "bad lua boolean value size found for key %s "
-                    "in shared_dict %s: %lu", key.data, name.data,
-                    (unsigned long) value.len);
+                              "in shared_dict %s: %lu", key.data, name.data,
+                              (unsigned long) value.len);
         }
 
         c = *value.data;
@@ -462,14 +502,29 @@ ngx_http_lua_shdict_get(lua_State *L)
         ngx_shmtx_unlock(&ctx->shpool->mutex);
 
         return luaL_error(L, "bad value type found for key %s in "
-                "shared_dict %s: %d", key.data, name.data,
-                value_type);
+                          "shared_dict %s: %d", key.data, name.data,
+                          value_type);
     }
 
     user_flags = sd->user_flags;
 
     ngx_shmtx_unlock(&ctx->shpool->mutex);
 
+    if (get_stale) {
+
+        /* always return value, flags, stale */
+
+        if (user_flags) {
+            lua_pushinteger(L, (lua_Integer) user_flags);
+
+        } else {
+            lua_pushnil(L);
+        }
+
+        lua_pushboolean(L, rc == NGX_DONE);
+        return 3;
+    }
+
     if (user_flags) {
         lua_pushinteger(L, (lua_Integer) user_flags);
         return 2;
@@ -488,7 +543,7 @@ ngx_http_lua_shdict_delete(lua_State *L)
 
     if (n != 2) {
         return luaL_error(L, "expecting 2 arguments, "
-                "but only seen %d", n);
+                          "but only seen %d", n);
     }
 
     lua_pushnil(L);
@@ -509,8 +564,7 @@ ngx_http_lua_shdict_flush_all(lua_State *L)
     n = lua_gettop(L);
 
     if (n != 1) {
-        return luaL_error(L, "expecting 1 argument, "
-                "but seen %d", n);
+        return luaL_error(L, "expecting 1 argument, but seen %d", n);
     }
 
     luaL_checktype(L, 1, LUA_TLIGHTUSERDATA);
@@ -557,8 +611,7 @@ ngx_http_lua_shdict_flush_expired(lua_State *L)
     n = lua_gettop(L);
 
     if (n != 1 && n != 2) {
-        return luaL_error(L, "expecting 1 or 2 argument(s), "
-                "but saw %d", n);
+        return luaL_error(L, "expecting 1 or 2 argument(s), but saw %d", n);
     }
 
     luaL_checktype(L, 1, LUA_TLIGHTUSERDATA);
@@ -569,7 +622,7 @@ ngx_http_lua_shdict_flush_expired(lua_State *L)
     }
 
     if (n == 2) {
-        attempts = luaL_checknumber(L, 2);
+        attempts = luaL_checkint(L, 2);
     }
 
     ctx = zone->data;
@@ -600,9 +653,7 @@ ngx_http_lua_shdict_flush_expired(lua_State *L)
                 ((u_char *) sd - offsetof(ngx_rbtree_node_t, color));
 
             ngx_rbtree_delete(&ctx->sh->rbtree, node);
-
             ngx_slab_free_locked(ctx->shpool, node);
-
             freed++;
 
             if (attempts && freed == attempts) {
@@ -652,7 +703,7 @@ ngx_http_lua_shdict_get_keys(lua_State *L)
     }
 
     if (n == 2) {
-        attempts = luaL_checknumber(L, 2);
+        attempts = luaL_checkint(L, 2);
     }
 
     ctx = zone->data;
@@ -725,6 +776,14 @@ ngx_http_lua_shdict_add(lua_State *L)
 }
 
 
+static int
+ngx_http_lua_shdict_safe_add(lua_State *L)
+{
+    return ngx_http_lua_shdict_set_helper(L, NGX_HTTP_LUA_SHDICT_ADD
+                                          |NGX_HTTP_LUA_SHDICT_SAFE_STORE);
+}
+
+
 static int
 ngx_http_lua_shdict_replace(lua_State *L)
 {
@@ -739,11 +798,17 @@ ngx_http_lua_shdict_set(lua_State *L)
 }
 
 
+static int
+ngx_http_lua_shdict_safe_set(lua_State *L)
+{
+    return ngx_http_lua_shdict_set_helper(L, NGX_HTTP_LUA_SHDICT_SAFE_STORE);
+}
+
+
 static int
 ngx_http_lua_shdict_set_helper(lua_State *L, int flags)
 {
     int                          i, n;
-    ngx_str_t                    name;
     ngx_str_t                    key;
     uint32_t                     hash;
     ngx_int_t                    rc;
@@ -751,7 +816,7 @@ ngx_http_lua_shdict_set_helper(lua_State *L, int flags)
     ngx_http_lua_shdict_node_t  *sd;
     ngx_str_t                    value;
     int                          value_type;
-    lua_Number                   num;
+    double                       num;
     u_char                       c;
     lua_Number                   exptime = 0;
     u_char                      *p;
@@ -767,30 +832,34 @@ ngx_http_lua_shdict_set_helper(lua_State *L, int flags)
 
     if (n != 3 && n != 4 && n != 5) {
         return luaL_error(L, "expecting 3, 4 or 5 arguments, "
-                "but only seen %d", n);
+                          "but only seen %d", n);
     }
 
-    luaL_checktype(L, 1, LUA_TLIGHTUSERDATA);
-
     zone = lua_touserdata(L, 1);
     if (zone == NULL) {
-        return luaL_error(L, "bad user data for the ngx_shm_zone_t pointer");
+        return luaL_error(L, "bad \"zone\" argument");
     }
 
     ctx = zone->data;
 
-    name = ctx->name;
+    if (lua_isnil(L, 2)) {
+        lua_pushnil(L);
+        lua_pushliteral(L, "nil key");
+        return 2;
+    }
 
     key.data = (u_char *) luaL_checklstring(L, 2, &key.len);
 
     if (key.len == 0) {
-        return luaL_error(L, "attempt to use empty keys");
+        lua_pushnil(L);
+        lua_pushliteral(L, "empty key");
+        return 2;
     }
 
     if (key.len > 65535) {
-        return luaL_error(L,
-                      "the key argument is more than 65535 bytes: %d",
-                      (int) key.len);
+        lua_pushnil(L);
+        lua_pushliteral(L, "key too long");
+        return 2;
     }
 
     hash = ngx_crc32_short(key.data, key.len);
@@ -803,7 +872,7 @@ ngx_http_lua_shdict_set_helper(lua_State *L, int flags)
         break;
 
     case LUA_TNUMBER:
-        value.len = sizeof(lua_Number);
+        value.len = sizeof(double);
         num = lua_tonumber(L, 3);
         value.data = (u_char *) &num;
         break;
@@ -816,7 +885,9 @@ ngx_http_lua_shdict_set_helper(lua_State *L, int flags)
 
     case LUA_TNIL:
         if (flags & (NGX_HTTP_LUA_SHDICT_ADD|NGX_HTTP_LUA_SHDICT_REPLACE)) {
-            return luaL_error(L, "attempt to add or replace nil values");
+            lua_pushnil(L);
+            lua_pushliteral(L, "attempt to add or replace nil values");
+            return 2;
         }
 
         value.len = 0;
@@ -824,9 +895,9 @@ ngx_http_lua_shdict_set_helper(lua_State *L, int flags)
         break;
 
     default:
-        return luaL_error(L, "unsupported value type for key \"%s\" in "
-                "shared_dict \"%s\": %s", key.data, name.data,
-                lua_typename(L, value_type));
+        lua_pushnil(L);
+        lua_pushliteral(L, "bad value type");
+        return 2;
     }
 
     if (n >= 4) {
@@ -840,8 +911,6 @@ ngx_http_lua_shdict_set_helper(lua_State *L, int flags)
         user_flags = (uint32_t) luaL_checkinteger(L, 5);
     }
 
-    dd("looking up key %s in shared dict %s", key.data, name.data);
-
     ngx_shmtx_lock(&ctx->shpool->mutex);
 
 #if 1
@@ -902,18 +971,18 @@ replace:
         if (value.data && value.len == (size_t) sd->value_len) {
 
             ngx_log_debug0(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
-                "lua shared dict set: found old entry and value size matched, "
-                "reusing it");
+                           "lua shared dict set: found old entry and value "
+                           "size matched, reusing it");
 
             ngx_queue_remove(&sd->queue);
             ngx_queue_insert_head(&ctx->sh->queue, &sd->queue);
 
-            sd->key_len = key.len;
+            sd->key_len = (u_short) key.len;
 
             if (exptime > 0) {
                 tp = ngx_timeofday();
                 sd->expires = (uint64_t) tp->sec * 1000 + tp->msec
-                        + exptime * 1000;
+                              + (uint64_t) (exptime * 1000);
 
             } else {
                 sd->expires = 0;
@@ -925,7 +994,7 @@ replace:
 
             dd("setting value type to %d", value_type);
 
-            sd->value_type = value_type;
+            sd->value_type = (uint8_t) value_type;
 
             p = ngx_copy(sd->data, key.data, key.len);
             ngx_memcpy(p, value.data, value.len);
@@ -939,8 +1008,8 @@ replace:
         }
 
         ngx_log_debug0(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
-            "lua shared dict set: found old entry bug value size NOT matched, "
-            "removing it first");
+                       "lua shared dict set: found old entry bug value size "
+                       "NOT matched, removing it first");
 
 remove:
         ngx_queue_remove(&sd->queue);
@@ -967,7 +1036,7 @@ insert:
     }
 
     ngx_log_debug0(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
-        "lua shared dict set: creating a new entry");
+                   "lua shared dict set: creating a new entry");
 
     n = offsetof(ngx_rbtree_node_t, color)
         + offsetof(ngx_http_lua_shdict_node_t, data)
@@ -978,9 +1047,17 @@ insert:
 
     if (node == NULL) {
 
+        if (flags & NGX_HTTP_LUA_SHDICT_SAFE_STORE) {
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+
+            lua_pushboolean(L, 0);
+            lua_pushliteral(L, "no memory");
+            return 2;
+        }
+
         ngx_log_debug1(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
-            "lua shared dict set: overriding non-expired items due to memory "
-            "shortage for entry \"%V\"", &name);
+                       "lua shared dict set: overriding non-expired items "
+                       "due to memory shortage for entry \"%V\"", &key);
 
         for (i = 0; i < 30; i++) {
             if (ngx_http_lua_shdict_expire(ctx, 0) == 0) {
@@ -1007,12 +1084,12 @@ allocated:
     sd = (ngx_http_lua_shdict_node_t *) &node->color;
 
     node->key = hash;
-    sd->key_len = key.len;
+    sd->key_len = (u_short) key.len;
 
     if (exptime > 0) {
         tp = ngx_timeofday();
         sd->expires = (uint64_t) tp->sec * 1000 + tp->msec
-                + exptime * 1000;
+                      + (uint64_t) (exptime * 1000);
 
     } else {
         sd->expires = 0;
@@ -1024,7 +1101,7 @@ allocated:
 
     dd("setting value type to %d", value_type);
 
-    sd->value_type = value_type;
+    sd->value_type = (uint8_t) value_type;
 
     p = ngx_copy(sd->data, key.data, key.len);
     ngx_memcpy(p, value.data, value.len);
@@ -1051,19 +1128,20 @@ ngx_http_lua_shdict_incr(lua_State *L)
     ngx_int_t                    rc;
     ngx_http_lua_shdict_ctx_t   *ctx;
     ngx_http_lua_shdict_node_t  *sd;
-    lua_Number                   num;
+    double                       num;
     u_char                      *p;
     ngx_shm_zone_t              *zone;
-    lua_Number                   value;
+    double                       value;
 
     n = lua_gettop(L);
 
     if (n != 3) {
-        return luaL_error(L, "expecting 3 arguments, "
-                "but only seen %d", n);
+        return luaL_error(L, "expecting 3 arguments, but only seen %d", n);
     }
 
-    luaL_checktype(L, 1, LUA_TLIGHTUSERDATA);
+    if (lua_type(L, 1) != LUA_TLIGHTUSERDATA) {
+        return luaL_error(L, "bad \"zone\" argument");
+    }
 
     zone = lua_touserdata(L, 1);
     if (zone == NULL) {
@@ -1072,16 +1150,24 @@ ngx_http_lua_shdict_incr(lua_State *L)
 
     ctx = zone->data;
 
+    if (lua_isnil(L, 2)) {
+        lua_pushnil(L);
+        lua_pushliteral(L, "nil key");
+        return 2;
+    }
+
     key.data = (u_char *) luaL_checklstring(L, 2, &key.len);
 
     if (key.len == 0) {
-        return luaL_error(L, "attempt to use empty keys");
+        lua_pushnil(L);
+        lua_pushliteral(L, "empty key");
+        return 2;
     }
 
     if (key.len > 65535) {
-        return luaL_error(L,
-                      "the key argument is more than 65535 bytes: %d",
-                      (int) key.len);
+        lua_pushnil(L);
+        lua_pushliteral(L, "key too long");
+        return 2;
     }
 
     hash = ngx_crc32_short(key.data, key.len);
@@ -1111,7 +1197,7 @@ ngx_http_lua_shdict_incr(lua_State *L)
 
     /* rc == NGX_OK */
 
-    if (sd->value_type != LUA_TNUMBER || sd->value_len != sizeof(lua_Number)) {
+    if (sd->value_type != LUA_TNUMBER || sd->value_len != sizeof(double)) {
         ngx_shmtx_unlock(&ctx->shpool->mutex);
 
         lua_pushnil(L);
@@ -1126,11 +1212,10 @@ ngx_http_lua_shdict_incr(lua_State *L)
 
     p = sd->data + key.len;
 
-    num = *(lua_Number *) p;
-
+    num = *(double *) p;
     num += value;
 
-    ngx_memcpy(p, (lua_Number *) &num, sizeof(lua_Number));
+    ngx_memcpy(p, (double *) &num, sizeof(double));
 
     ngx_shmtx_unlock(&ctx->shpool->mutex);
 
@@ -1201,7 +1286,7 @@ ngx_http_lua_shared_dict_get(ngx_shm_zone_t *zone, u_char *key_data,
 
     case LUA_TNUMBER:
 
-        if (len != sizeof(lua_Number)) {
+        if (len != sizeof(double)) {
             ngx_log_error(NGX_LOG_ERR, ngx_cycle->log, 0, "bad lua number "
                           "value size found for key %*s: %lu", key_len,
                           key_data, (unsigned long) len);
@@ -1258,6 +1343,7 @@ ngx_http_lua_find_zone(u_char *name_data, size_t name_len)
             if (part->next == NULL) {
                 break;
             }
+
             part = part->next;
             zone = part->elts;
             i = 0;
@@ -1278,3 +1364,437 @@ ngx_http_lua_find_zone(u_char *name_data, size_t name_len)
     return NULL;
 }
 
+
+#ifndef NGX_HTTP_LUA_NO_FFI_API
+int
+ngx_http_lua_ffi_shdict_store(ngx_shm_zone_t *zone, int op, u_char *key,
+    size_t key_len, int value_type, u_char *str_value_buf,
+    size_t str_value_len, double num_value, int exptime, int user_flags,
+    char **errmsg, int *forcible)
+{
+    int                          i, n;
+    u_char                       c, *p;
+    uint32_t                     hash;
+    ngx_int_t                    rc;
+    ngx_time_t                  *tp;
+    ngx_rbtree_node_t           *node;
+    ngx_http_lua_shdict_ctx_t   *ctx;
+    ngx_http_lua_shdict_node_t  *sd;
+
+    if (zone == NULL) {
+        return NGX_ERROR;
+    }
+
+    dd("exptime: %d", exptime);
+
+    ctx = zone->data;
+
+    *forcible = 0;
+
+    hash = ngx_crc32_short(key, key_len);
+
+    switch (value_type) {
+    case LUA_TSTRING:
+        /* do nothing */
+        break;
+
+    case LUA_TNUMBER:
+        dd("num value: %lf", num_value);
+        str_value_buf = (u_char *) &num_value;
+        str_value_len = sizeof(double);
+        break;
+
+    case LUA_TBOOLEAN:
+        c = num_value ? 1 : 0;
+        str_value_buf = &c;
+        str_value_len = sizeof(u_char);
+        break;
+
+    case LUA_TNIL:
+        if (op & (NGX_HTTP_LUA_SHDICT_ADD|NGX_HTTP_LUA_SHDICT_REPLACE)) {
+            *errmsg = "attempt to add or replace nil values";
+            return NGX_ERROR;
+        }
+
+        str_value_buf = NULL;
+        str_value_len = 0;
+        break;
+
+    default:
+        *errmsg = "unsupported value type";
+        return NGX_ERROR;
+    }
+
+    ngx_shmtx_lock(&ctx->shpool->mutex);
+
+#if 1
+    ngx_http_lua_shdict_expire(ctx, 1);
+#endif
+
+    rc = ngx_http_lua_shdict_lookup(zone, hash, key, key_len, &sd);
+
+    dd("lookup returns %d", (int) rc);
+
+    if (op & NGX_HTTP_LUA_SHDICT_REPLACE) {
+
+        if (rc == NGX_DECLINED || rc == NGX_DONE) {
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+            *errmsg = "not found";
+            return NGX_DECLINED;
+        }
+        /* rc == NGX_OK */
+
+        goto replace;
+    }
+
+    if (op & NGX_HTTP_LUA_SHDICT_ADD) {
+
+        if (rc == NGX_OK) {
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+            *errmsg = "exists";
+            return NGX_DECLINED;
+        }
+
+        if (rc == NGX_DONE) {
+            /* exists but expired */
+
+            dd("go to replace");
+            goto replace;
+        }
+
+        /* rc == NGX_DECLINED */
+
+        dd("go to insert");
+        goto insert;
+    }
+
+    if (rc == NGX_OK || rc == NGX_DONE) {
+
+        if (value_type == LUA_TNIL) {
+            goto remove;
+        }
+
+replace:
+        if (str_value_buf && str_value_len == (size_t) sd->value_len) {
+
+            ngx_log_debug0(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
+                           "lua shared dict set: found old entry and value "
+                           "size matched, reusing it");
+
+            ngx_queue_remove(&sd->queue);
+            ngx_queue_insert_head(&ctx->sh->queue, &sd->queue);
+
+            sd->key_len = (u_short) key_len;
+
+            if (exptime > 0) {
+                tp = ngx_timeofday();
+                sd->expires = (uint64_t) tp->sec * 1000 + tp->msec
+                              + (uint64_t) exptime;
+
+            } else {
+                sd->expires = 0;
+            }
+
+            sd->user_flags = user_flags;
+
+            sd->value_len = (uint32_t) str_value_len;
+
+            dd("setting value type to %d", value_type);
+
+            sd->value_type = (uint8_t) value_type;
+
+            p = ngx_copy(sd->data, key, key_len);
+            ngx_memcpy(p, str_value_buf, str_value_len);
+
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+
+            return NGX_OK;
+        }
+
+        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
+                       "lua shared dict set: found old entry bug value size "
+                       "NOT matched, removing it first");
+
+remove:
+        ngx_queue_remove(&sd->queue);
+
+        node = (ngx_rbtree_node_t *)
+                   ((u_char *) sd - offsetof(ngx_rbtree_node_t, color));
+
+        ngx_rbtree_delete(&ctx->sh->rbtree, node);
+
+        ngx_slab_free_locked(ctx->shpool, node);
+
+    }
+
+insert:
+    /* rc == NGX_DECLINED or value size unmatch */
+
+    if (str_value_buf == NULL) {
+        ngx_shmtx_unlock(&ctx->shpool->mutex);
+        return NGX_OK;
+    }
+
+    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
+                   "lua shared dict set: creating a new entry");
+
+    n = offsetof(ngx_rbtree_node_t, color)
+        + offsetof(ngx_http_lua_shdict_node_t, data)
+        + key_len
+        + str_value_len;
+
+    node = ngx_slab_alloc_locked(ctx->shpool, n);
+
+    if (node == NULL) {
+
+        if (op & NGX_HTTP_LUA_SHDICT_SAFE_STORE) {
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+
+            *errmsg = "no memory";
+            return NGX_ERROR;
+        }
+
+        ngx_log_debug2(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
+                       "lua shared dict set: overriding non-expired items "
+                       "due to memory shortage for entry \"%*s\"", key_len,
+                       key);
+
+        for (i = 0; i < 30; i++) {
+            if (ngx_http_lua_shdict_expire(ctx, 0) == 0) {
+                break;
+            }
+
+            *forcible = 1;
+
+            node = ngx_slab_alloc_locked(ctx->shpool, n);
+            if (node != NULL) {
+                goto allocated;
+            }
+        }
+
+        ngx_shmtx_unlock(&ctx->shpool->mutex);
+
+        *errmsg = "no memory";
+        return NGX_ERROR;
+    }
+
+allocated:
+    sd = (ngx_http_lua_shdict_node_t *) &node->color;
+
+    node->key = hash;
+    sd->key_len = (u_short) key_len;
+
+    if (exptime > 0) {
+        tp = ngx_timeofday();
+        sd->expires = (uint64_t) tp->sec * 1000 + tp->msec
+                      + (uint64_t) exptime;
+
+    } else {
+        sd->expires = 0;
+    }
+
+    sd->user_flags = user_flags;
+    sd->value_len = (uint32_t) str_value_len;
+    dd("setting value type to %d", value_type);
+    sd->value_type = (uint8_t) value_type;
+
+    p = ngx_copy(sd->data, key, key_len);
+    ngx_memcpy(p, str_value_buf, str_value_len);
+
+    ngx_rbtree_insert(&ctx->sh->rbtree, node);
+    ngx_queue_insert_head(&ctx->sh->queue, &sd->queue);
+    ngx_shmtx_unlock(&ctx->shpool->mutex);
+
+    return NGX_OK;
+}
+
+
+int
+ngx_http_lua_ffi_shdict_get(ngx_shm_zone_t *zone, u_char *key,
+    size_t key_len, int *value_type, u_char **str_value_buf,
+    size_t *str_value_len, double *num_value, int *user_flags,
+    int get_stale, int *is_stale)
+{
+    ngx_str_t                    name;
+    uint32_t                     hash;
+    ngx_int_t                    rc;
+    ngx_http_lua_shdict_ctx_t   *ctx;
+    ngx_http_lua_shdict_node_t  *sd;
+    ngx_str_t                    value;
+
+    if (zone == NULL) {
+        return NGX_ERROR;
+    }
+
+    ctx = zone->data;
+    name = ctx->name;
+
+    hash = ngx_crc32_short(key, key_len);
+
+#if (NGX_DEBUG)
+    ngx_log_debug3(NGX_LOG_DEBUG_HTTP, ctx->log, 0,
+                   "fetching key \"%*s\" in shared dict \"%V\"", key_len,
+                   key, &name);
+#endif /* NGX_DEBUG */
+
+    ngx_shmtx_lock(&ctx->shpool->mutex);
+
+#if 1
+    if (!get_stale) {
+        ngx_http_lua_shdict_expire(ctx, 1);
+    }
+#endif
+
+    rc = ngx_http_lua_shdict_lookup(zone, hash, key, key_len, &sd);
+
+    dd("shdict lookup returns %d", (int) rc);
+
+    if (rc == NGX_DECLINED || (rc == NGX_DONE && !get_stale)) {
+        ngx_shmtx_unlock(&ctx->shpool->mutex);
+        *value_type = LUA_TNIL;
+        return NGX_OK;
+    }
+
+    /* rc == NGX_OK || (rc == NGX_DONE && get_stale) */
+
+    *value_type = sd->value_type;
+
+    dd("data: %p", sd->data);
+    dd("key len: %d", (int) sd->key_len);
+
+    value.data = sd->data + sd->key_len;
+    value.len = (size_t) sd->value_len;
+
+    if (*str_value_len < (size_t) value.len) {
+        if (*value_type != LUA_TSTRING) {
+            return NGX_ERROR;
+        }
+
+        *str_value_buf = malloc(value.len);
+        if (*str_value_buf == NULL) {
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+            return NGX_ERROR;
+        }
+    }
+
+    *str_value_len = value.len;
+
+    switch (*value_type) {
+    case LUA_TSTRING:
+        ngx_memcpy(*str_value_buf, value.data, value.len);
+        break;
+
+    case LUA_TNUMBER:
+
+        if (value.len != sizeof(double)) {
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+            ngx_log_error(NGX_LOG_ERR, ngx_cycle->log, 0,
+                          "bad lua number value size found for key %*s "
+                          "in shared_dict %V: %z", key_len, key,
+                          &name, value.len);
+            return NGX_ERROR;
+        }
+
+        *num_value = *(double *) value.data;
+        break;
+
+    case LUA_TBOOLEAN:
+
+        if (value.len != sizeof(u_char)) {
+            ngx_shmtx_unlock(&ctx->shpool->mutex);
+            ngx_log_error(NGX_LOG_ERR, ngx_cycle->log, 0,
+                          "bad lua boolean value size found for key %*s "
+                          "in shared_dict %V: %z", key_len, key, &name,
+                          value.len);
+            return NGX_ERROR;
+        }
+
+        ngx_memcpy(*str_value_buf, value.data, value.len);
+        break;
+
+    default:
+
+        ngx_shmtx_unlock(&ctx->shpool->mutex);
+        ngx_log_error(NGX_LOG_ERR, ngx_cycle->log, 0,
+                      "bad value type found for key %*s in "
+                      "shared_dict %V: %d", key_len, key, &name,
+                      *value_type);
+    }
+
+    *user_flags = sd->user_flags;
+    dd("user flags: %d", *user_flags);
+
+    ngx_shmtx_unlock(&ctx->shpool->mutex);
+
+    if (get_stale) {
+
+        /* always return value, flags, stale */
+
+        *is_stale = (rc == NGX_DONE);
+        return NGX_OK;
+    }
+
+    return NGX_OK;
+}
+
+
+int
+ngx_http_lua_ffi_shdict_incr(ngx_shm_zone_t *zone, u_char *key,
+    size_t key_len, double *value, char **err)
+{
+    uint32_t                     hash;
+    ngx_int_t                    rc;
+    ngx_http_lua_shdict_ctx_t   *ctx;
+    ngx_http_lua_shdict_node_t  *sd;
+    double                       num;
+    u_char                      *p;
+
+    ctx = zone->data;
+    hash = ngx_crc32_short(key, key_len);
+
+    dd("looking up key %.*s in shared dict %.*s", (int) key_len, key,
+       (int) ctx->name.len, ctx->name.data);
+
+    ngx_shmtx_lock(&ctx->shpool->mutex);
+#if 1
+    ngx_http_lua_shdict_expire(ctx, 1);
+#endif
+    rc = ngx_http_lua_shdict_lookup(zone, hash, key, key_len, &sd);
+
+    dd("shdict lookup returned %d", (int) rc);
+
+    if (rc == NGX_DECLINED || rc == NGX_DONE) {
+        ngx_shmtx_unlock(&ctx->shpool->mutex);
+        *err = "not found";
+        return NGX_ERROR;
+    }
+
+    /* rc == NGX_OK */
+
+    if (sd->value_type != LUA_TNUMBER || sd->value_len != sizeof(double)) {
+        ngx_shmtx_unlock(&ctx->shpool->mutex);
+        *err = "not a number";
+        return NGX_ERROR;
+    }
+
+    ngx_queue_remove(&sd->queue);
+    ngx_queue_insert_head(&ctx->sh->queue, &sd->queue);
+
+    dd("setting value type to %d", (int) sd->value_type);
+
+    p = sd->data + key_len;
+
+    num = *(double *) p;
+    num += *value;
+
+    ngx_memcpy(p, (double *) &num, sizeof(double));
+
+    ngx_shmtx_unlock(&ctx->shpool->mutex);
+
+    *value = num;
+    return NGX_OK;
+}
+#endif /* NGX_HTTP_LUA_NO_FFI_API */
+
+
+/* vi:set ft=c ts=4 sw=4 et fdm=marker: */
