@@ -1,5 +1,11 @@
-#ifndef NGX_HTTP_LUA_SHDICT_H
-#define NGX_HTTP_LUA_SHDICT_H
+
+/*
+ * Copyright (C) Yichun Zhang (agentzh)
+ */
+
+
+#ifndef _NGX_HTTP_LUA_SHDICT_H_INCLUDED_
+#define _NGX_HTTP_LUA_SHDICT_H_INCLUDED_
 
 
 #include "ngx_http_lua_common.h"
@@ -35,13 +41,12 @@ typedef struct {
 
 
 ngx_int_t ngx_http_lua_shdict_init_zone(ngx_shm_zone_t *shm_zone, void *data);
-
 void ngx_http_lua_shdict_rbtree_insert_value(ngx_rbtree_node_t *temp,
     ngx_rbtree_node_t *node, ngx_rbtree_node_t *sentinel);
-
 void ngx_http_lua_inject_shdict_api(ngx_http_lua_main_conf_t *lmcf,
-        lua_State *L);
+    lua_State *L);
 
 
-#endif /* NGX_HTTP_LUA_SHDICT_H */
+#endif /* _NGX_HTTP_LUA_SHDICT_H_INCLUDED_ */
 
+/* vi:set ft=c ts=4 sw=4 et fdm=marker: */
